from mwmatching import maxWeightMatching

from copy import copy
import os
import pickle

import networkx as nx
import sys

from string_align import sequence_align

# Existen los ficheros para almacenar la info?
if not os.path.isfile("compscore.pkl"):
    # Creamos el fichero
    f = open("compscore.pkl", "wb")
    pickle.dump({}, f)
    f.close()
if not os.path.isfile("enzscore.pkl"):
    # Creamos el fichero
    f = open("enzscore.pkl", "wb")
    pickle.dump({}, f)
    f.close()

class Align:
    def __init__(self, route1, route2):
        # Data form pathway
        self.route1 = route1
        self.route2 = route2
 
        # Compare
        self.gap_pen = 0
        self.compscore = {} # Compound score
        self.enzymscore = {} # Enzyme score
        
        self._load_score()
        
        self._rscore = {}
        
        # Penalizacion para evitar
        # que componentes diferentes tengan 
        # similitud 1.
        self._max_score_compound = 0.99

        # Align
        self._align()
        self._path_align = None
        self._reaction_align = None
        
    def path_align(self):
        """
        Devuelve el mejor alineamiento de caminos.
        reacion_set() --> {[i,j]: score}
            Devuelve el score al comparar 
                route1.reaction_set()[i] 
                    con
                route2.reaction_set()[j]
        """
        if not self._path_align is None:
            return self._path_align
        
        graph = []
        sscore = {}
        
        set1 = self.route1.reaction_set()
        set2 = self.route2.reaction_set()
        len1 = len(set1)
        
        self._max_path = max(len(set1), len(set2))
        
        for index1 in xrange(len(set1)):
            path1 = set1[index1]
            for index2 in xrange(len(set2)):
                path2 = set2[index2]
                tmp = self._compare_path(path1, path2)
                graph.append([index1, index2+len1, tmp[2]])
                sscore[(index1, index2+len1)] = tmp
                
        # Alineamiento por
        match = maxWeightMatching(graph, maxcardinality=True)
        self._path_align = {}
        for i in xrange(len1):
            if match[i] != -1:
                self._path_align[(i, match[i]-len1)] = sscore[(i,match[i])]
        
        return self._path_align
    
    def reaction_align(self):
        if not self._reaction_align is None:
            return self._reaction_align
        
        score = self.path_align()
        repeticiones = {}
        graph = []
        ruta1 = self.route1.reactions()
        ruta2 = self.route2.reactions()
        N1 = len(ruta1)
        
        for match in score:
            # Alineamiento reaccion match[0] <--> match[1]
            path1 = score[match][0]
            path2 = score[match][1]
            
            for i in xrange(len(path1)):
                if not path1[i] in repeticiones:
                    repeticiones[path1[i]] = {}
                if not path2[i] in repeticiones[path1[i]]:
                    repeticiones[path1[i]][path2[i]] = [0, 0.0, 0.0]
                
                repeticiones[path1[i]][path2[i]][0] = repeticiones[path1[i]][path2[i]][0] + 1
                repeticiones[path1[i]][path2[i]][1] = repeticiones[path1[i]][path2[i]][1] + score[match][2]
                if repeticiones[path1[i]][path2[i]][2] < score[match][2]:
                    repeticiones[path1[i]][path2[i]][2] = score[match][2]

        for react1 in repeticiones:
            for react2 in repeticiones[react1]:
                if react1 != "-" and react2 != "-":
                    ii = ruta1.index(react1)
                    jj = N1 + ruta2.index(react2)
                
                    graph.append([ii,jj,repeticiones[react1][react2][0]])
            
        match = maxWeightMatching(graph, maxcardinality=True) 

        self._reaction_align = {}

        for i in xrange(0, N1):
            if match[i] == -1:
                self._reaction_align[ruta1[i]] = [None, 0]
            else:
                jj = ruta1[i]
                mm = ruta2[match[i]-N1]
                # jj <--> mm
                #REP = repeticiones[jj][mm][0]
                #MEDIA = 1.0*repeticiones[jj][mm][1]/REP
                #MEDIA1 = MEDIA/len(repeticiones[jj])
                #if repeticiones[jj][mm][2] != 0:
                #    MEDIA2 = MEDIA/repeticiones[jj][mm][2]
                #else:
                #    MEDIA2 = 0
                MAX = 1.0*repeticiones[jj][mm][2]
                #REPPERC = 1.0*REP/sum([repeticiones[jj][ll][0] for ll in repeticiones[jj]])
                #self._reaction_align[ruta1[i]] = [mm, mm, MEDIA1, MEDIA2, MEDIA, REPPERC, REPPERC*MAX, MAX]
                self._reaction_align[ruta1[i]] = [mm, MAX]

        return self._reaction_align

    def motiff(self):
        ralign = self.reaction_align()
        react1 = self.route1.reactions()
        react2 = self.route2.reactions()

        graph = nx.Graph()
        graph.add_nodes_from(react1)
        for r1 in react1:
            for r2 in react2:
                exit1 = set(self.route1.reaction_to(r1))
                start1 = set(self.route1.reaction_from(r2))
                if len( exit1.intersection(start1) ) > 0:
                    if r1 in ralign and r2 in ralign:
                        rr1 = ralign[r1][0]
                        rr2 = ralign[r2][0]
                        exit1 = set(self.route2.reaction_to(rr1))
                        start1 = set(self.route2.reaction_from(rr2))
                        if len( exit1.intersection(start1) ) > 0:
                            graph.add_edge(r1, r2)

        componente = nx.connected_components(graph)[0]
        resultado = [componente[0]]
        tmp = [componente[0]]
        yavistos = {componente[0]: True}

        while len(tmp) > 0:
            tmpN = tmp.pop()
            tmpL = graph[tmpN].keys()
            for a in tmpL:
                if a not in yavistos:
                    yavistos[a] = True
                    tmp[0:0] = graph[a].keys()
                    resultado.append(a)

        return resultado
    
    def path_score(self):
        tmp = self.path_align()
        score = 0.0
        for i in tmp:
            score = score + tmp[i][2]
        if self._max_path == 0:
            return 0
        return score/self._max_path
    
    def reaction_score(self):
        MAX = max(len(self.route1.reactions()), len(self.route2.reactions()))
        tmp = self.reaction_align()
        score1 = 0.0
        if MAX == 0:
            return 0.0
        for i in tmp:
            tmp1 = tmp[i][1]
            score1 = score1 + tmp1
        return score1/MAX
            
    def final_score(self):
        return self.path_score(), self.reaction_score()
        
    def _align(self):
        graph = []
        self.reactions1 = self.route1.reactions()
        self.reactions2 = self.route2.reactions()
        len1 = len(self.reactions1)
        for react1 in self.route1.reactions():
            index1 = self.reactions1.index(react1)
            for react2 in self.route2.reactions():
                index2 = self.reactions2.index(react2)
                graph.append([index1, len1 + index2, 
                              self._compare_reactions(react1, react2)])

    def _load_score(self):
        f = open("compscore.pkl", "rb")
        self.compscore = pickle.load(f)
        f.close()
        
        f = open("enzscore.pkl", "rb")
        self.enzymscore = pickle.load(f)
        f.close()
        
    def _update(self, f = "both"):
        if f == "both" or f == "compound":
            f = open("compscore.pkl", "wb")
            pickle.dump(self.compscore, f)
            f.close()
            
        if f == "both" or f == "enzym":
            f = open("enzscore.pkl", "wb")
            pickle.dump(self.enzymscore, f)
            f.close()

    def _compare_enzyme(self, enzyme1, enzyme2):
        # Already computed?
        if (enzyme1, enzyme2) in self.enzymscore:
            return self.enzymscore[(enzyme1, enzyme2)]
        elif (enzyme2, enzyme1) in self.enzymscore:
            return self.enzymscore[(enzyme2, enzyme1)]

        # if not, then compute it
        tmp1 = enzyme1.split(":")[1].split(".") #ec:1.1.1.1 --> [1, 1, 1, 1]
        tmp2 = enzyme2.split(":")[1].split(".")
    
        ret = 0.0
        if tmp1 == tmp2:
            ret = 1.0
        elif tmp1[0:2] == tmp2[0:2]:
            ret = 0.75
        elif tmp1[0:1] == tmp2[0:1]:
            ret = 0.5
        elif tmp1[0] == tmp2[0]:
            ret = 0.25

        self.enzymscore[(enzyme1, enzyme2)] = ret #save
        
        self._update("enzym")
        
        return ret
    
    def _compare_compound(self, compound1, compound2):
        # cpd:C00100-12 --> C00100-12 --> C00100 --> 00100
        compound1 = compound1.split(":")[1].split("-")[0][1:]
        compound2 = compound2.split(":")[1].split("-")[0][1:]
        
        # Already computed?
        penalitzacio = 1.0
        if compound1 != compound2:
            penalitzacio = self._max_score_compound
        if (compound1, compound2) in self.compscore:
            return penalitzacio*self.compscore[(compound1, compound2)]
        elif (compound2, compound1) in self.compscore:
            return penalitzacio*self.compscore[(compound2, compound1)]

        # if not, the compute it
        ret = 0.0
    
        if compound1 == compound2:
            ret = 1.0
        else:  
            try:
                ret = float(os.popen("./simcomp compound2 -e " + compound1 + \
                                             " -f " + compound2 + \
                                             " | awk {'print $6'}").read())
            except:
                print "I can't compute the score between '" + str(compound1) + "' and '" + str(compound2) + "'\n"
                print "Maybe ./simcomp can't be executed \n"
                print "or one of the components isn't in the database 'compound2'"
                sys.exit(0)

        self.compscore[(compound1, compound2)] = ret #save
        self._update("compound")
        
        return ret

    def _compare_set_compound(self, set1, set2):
        bi_graph = []
        len1 = len(set1)
        for cpd1 in set1:
            index1 = set1.index(cpd1)
            for cpd2 in set2:
                bi_graph.append([index1, len1 + set2.index(cpd2), 
                                 self._compare_compound(cpd1, cpd2)])
                
        match = maxWeightMatching(bi_graph)
        score = 0.0
        i = 0
        while (i < len(match)) and (i < min(len1, len(set2))):
            if match[i] != -1:
                score = score + self._compare_compound(set1[i], 
                                                       set2[match[i]-len1])
            i = i+1
        
        if max(len(set1), len(set2)) == 0:
            return 0.0
        return score/max(len(set1), len(set2))

    def _compare_reactions(self, reaction1, reaction2):
        if reaction1 == reaction2:
            return 1.0
        if (reaction1, reaction2) in self._rscore:
            return self._rscore[(reaction1, reaction2)]

        # Entrada
        input1 = self.route1.reaction_from(reaction1)
        input2 = self.route2.reaction_from(reaction2)
        
        iscore = self._compare_set_compound(input1, input2)
        
        # Salida
        input1 = self.route1.reaction_to(reaction1)
        input2 = self.route2.reaction_to(reaction2)
        
        oscore = self._compare_set_compound(input1, input2)
        
        escore= self._compare_enzyme(self.route1.enzyme(reaction1), 
                                     self.route2.enzyme(reaction2))
        
        self._rscore[(reaction1, reaction2)] = escore * 0.4 + iscore * 0.3 + oscore * 0.3
        self._rscore[(reaction2, reaction1)] = escore * 0.4 + iscore * 0.3 + oscore * 0.3
        return escore * 0.4 + iscore * 0.3 + oscore * 0.3
    
    def _compare_path(self, path1, path2):
        if path1 == path2:
            return [path1, path2, 1.0]
        return sequence_align(path1, path2, 
                               self._compare_reactions, self.gap_pen)
        
