from copy import copy
#
# Este fichero contiene la especificacion de 
# las clases que deben formar el paquete.
#

class NotImplementedException(Exception):
    pass

class NoInfoAvailableException(Exception):
    pass

class Route:
    def components(self):
        """
        Debe devolver una lista con los componentes que forman la ruta.
        """
        raise NotImplementedException("Base class")

    def reactions(self):
        """
        Debe devolver una lista con las reaciones que forman la ruta.
        """
        raise NotImplementedException("Base class")

    def from_component(self, component):
        """
        Debe devolver la lista de reacciones que salen de un componente dado.
        """
        raise NotImplementedException("Base class")

    def reaction_to(self, reaction):
        """
        Debe devolver la llegada de la reaccion
        """
        raise NotImplementedException("Base class")

    def enzyme(self, reaction):
        """
        Debe devolver la lista de encimas que hacen posible la reaccion.
        """
        raise NotImplementedException("Base class")

    #######################################################
    #              IMPLEMENTATION                         #
    #######################################################
    
    def _clean_visited(self):
        for node in self.components():
            self._visitats[node] = False
        for arista in self.reactions():
            self._visitats[arista] = False

    def _check_len(self, l):
        # Comprovam que sigui el cami mes llarg
        existent = False
        for i in self._result:
            if set(l).issubset(set(i)):
                existent = True
                break
            if set(l).issuperset(set(i)):
                self._result[self._result.index(i)] = copy(l)
                existent = True
                break
        if not existent:
            self._result.append(copy(l))

    def _calc_reaction(self, l, x):
        self._visitats[x] = True

        # Por donde podemos seguir el camino?
        for reaction in self.from_component(x):
            if not self._visitats[reaction]:
                # No hemos pasado por aqui
                self._visitats[reaction] = True

                # Backtracking
                l.append(reaction)
                for y in self.reaction_to(reaction):
                    if not self._visitats[y]:
                        self._calc_reaction(l, y)
                self._check_len(l[:])

                l.remove(reaction)

    def reaction_set(self):
        self._result = []
        self._visitats = {}

        self._clean_visited()

        for node in self.components():
            tmp = []
            self._calc_reaction(tmp, node)
            self._clean_visited()
        
        # Add length 1 path
        tmp = []
        for path in self._result:
            tmp[0:0] = path
        
        for reaction in self.reactions():
            if reaction not in tmp:
                self._result.append([reaction])

        return self._result

    def reaction_from(self, reaction):
        tmp = []
        for component in self.components():
            if reaction in self.from_component(component):
                tmp.append(component)
        return tmp
    
class DynamicRoute(Route):
    def dynamic_info(self, reaction):
        """
        Informacion dinamica de la reaccion
        """
        raise NotImplementedException("Base class")
