from base import NoInfoAvailableException, Route

from SOAPpy import WSDL

class KEGGRoute(Route):
  def __init__(self, pathway, debug = False):
    self.kegg = WSDL.Proxy('http://soap.genome.jp/KEGG.wsdl')

    self._components = {}
    self._reactions = {}

    if debug:
      print "Retrieving info from KEGG."
      print "Downloading set of compunds."

    tmp_components = set(self.kegg.get_compounds_by_pathway(pathway))
    if len(tmp_components) == 0:
      raise NoInfoAvailableException("No info retrieved from KEGG. Maybe the pathway doesn't exist.")

    if debug:
      print "Downloading set of reactions."
    tmp_reactions = set(self.kegg.get_reactions_by_pathway(pathway))

    if debug:
      print "Downloading info. between compounds and reactions."

    for component in tmp_components:
      specific_reactions = set(self.kegg.get_reactions_by_compound(component))
      self._components[component] = list(specific_reactions.intersection(tmp_reactions))

  def components(self):
    return self._components.keys()

  def from_component(self, component):
    if component in self._components:
      return self._components[component]
    else:
      return []
    
