from base import NoInfoAvailableException, Route

import xml.parsers.expat

class PNMLParser(Route):
  def _start_element(self, name, attrs):
    # XML Parser
    if name == "place":
      self._components[attrs.values()[0]] = []
    elif name == "transition":
      self._reactions[attrs.values()[0]] = []
      self._exit_reaction[attrs.values()[0]] = []
    elif name == "arc":
      if attrs['source'][0:3] == "cpd" and attrs['source'] not in self._components:
        if attrs['target'] not in self._components:
          print "ADD", attrs['source']
          self._components[attrs['target']] = []
      if (not attrs['source'][0:3] == "cpd") and attrs['source'] not in self._reactions:
        if attrs['target'] not in self._components:
          print "ADD", attrs['target']
          self._reactions[attrs['target']] = []

      if attrs['source'] in self._components:
        self._components[attrs['source']].append(attrs['target'])
        self._exit_reaction[attrs['target']].append(attrs['source'])
        if attrs['target'] not in self._reactions:
          self._reactions[attrs['target']] = []
      elif attrs['source'] in self._reactions:
        self._reactions[attrs['source']].append(attrs['target'])
        if attrs['target'] not in self._components:
          self._components[attrs['target']] = []

  def __init__(self, f):
    # component -> reactions starting in component
    self._components = {}
    # reaction -> arriving component
    self._reactions = {}
    # from which component --> reaction
    self._exit_reaction = {}

    self.enz_db = None

    # Parsing PNML file
    parser = xml.parsers.expat.ParserCreate()
    parser.StartElementHandler = self._start_element
    parser.Parse(open(f).readlines()[0])
    
  def components(self):
    return self._components.keys()

  def reactions(self):
    return self._reactions.keys()

  def exit_reactions(self):
    return self._exit_reaction.keys()

  def from_component(self, component):
    if component in self._components:
      return self._components[component]
    else:
      return []

  def reaction_to(self, reaction):
    if reaction in self._reactions:
      return self._reactions[reaction]
    else:
      return []

  def reaction_from(self, reaction):
    if reaction in self._exit_reaction:
      return self._exit_reaction[reaction]
    else:
      return []

  def enzyme_info(self, parser):
    self.enz_db = parser

  def enzyme(self, reaction):
    if self.enz_db is None:
      raise NoInfoAvailableException("No info about enzyme found")
    else:
      return self.enz_db.get_enzyme(reaction)

class PNMLEnzymeParser:
  def _start_element(self, name, attrs):
    # XML Parser
    #print name
    if name == "entry":
      #print attrs["type"]
      if attrs["type"] == "enzyme":
        self.id_to_enzyme[attrs["id"]] = attrs["name"]

  def __init__(self, f):
    self.id_to_enzyme = {}

    parser = xml.parsers.expat.ParserCreate()
    parser.StartElementHandler = self._start_element
    parser.Parse("".join(open(f).readlines()))

  def get_enzyme(self, reaction):
    reaction = reaction.split("#")[0]
    if "-" in reaction:
      tmp = reaction.split("-")
      if tmp[-1] in self.id_to_enzyme:
        return self.id_to_enzyme[tmp[-1]]
      else:
        #raise NoInfoAvailableException("ID " + str(tmp[-1]) + " not found in EC file.")
        return "ec:0.0.0.0"
    else:
      if str(reaction) in self.id_to_enzyme:
        return self.id_to_enzyme[str(reaction)]
      else:
        #raise NoInfoAvailableException("ID " + str(reaction) + " not found in EC file.")
        return "ec:0.0.0.0"
